@@ -87,15 +87,15 @@ class MfgBase(object):
 
     # final condition
     def UT_func(self, x):
-        raise NotImplementedError(self.__class__.__name__ + "didn't implement M0_func")
+        raise NotImplementedError(self.__class__.__name__ + "didn't implement UT_func")
 
     # hamilton
     # H(Du,m) = 0.5*|Du|^2/F1(m) - F2(m)
     def F1_func(self, m):
-        raise NotImplementedError(self.__class__.__name__ + "didn't implement M0_func")
+        raise NotImplementedError(self.__class__.__name__ + "didn't implement F1_func")
 
     def F2_func(self, m):
-        raise NotImplementedError(self.__class__.__name__ + "didn't implement M0_func")
+        raise NotImplementedError(self.__class__.__name__ + "didn't implement F2_func")
 
     def hjb_rhs(self, un, q, mn):
         """
@@ -210,13 +210,16 @@ class MfgBase(object):
         return
 
     def default_fp_pn(n):
-        return (1 - 1 / n, 1 / n)
+        if n == 0:
+            return (0, 1)
+        else:
+            return (1 - 1 / (n+1), 1 / (n+1))
 
     def solve(self, maxit=10, fp=False, pn=default_fp_pn):
         # fp: bool, use fictious play or not
         self.diff_M = []
         self.diff_M_Mbar = []
-        for i in range(1, maxit + 1):
+        for i in range(0, maxit):
             # solve FP with Q
             Q = (self.QL, self.QR)
             self.solve_FP(Q)
@@ -241,15 +244,15 @@ class MfgBase(object):
             self.logging_solution(self.M, self.U)
         return
 
-    def default_fp1_pn(n):
-        return (n / (n + 1), 1 / (n + 1))
+    def default_spi1_pn(n):
+        return (n / (n + 2), 1 / (n + 2))
 
-    def solve_with_fp1(self, maxit=10, pn=default_fp1_pn):
+    def solve_with_spi1(self, maxit=10, pn=default_spi1_pn):
         self.diff_M = []
         # init
         QL_bar = np.copy(self.QL)
         QR_bar = np.copy(self.QR)
-        for i in range(1, maxit + 1):
+        for i in range(0, maxit):
             # solve FP with Q_bar
             Q_bar = (QL_bar, QR_bar)
             self.solve_FP(Q_bar)
@@ -269,10 +272,13 @@ class MfgBase(object):
             self.logging_solution(self.M, self.U)
         return
 
-    def default_fp2_pn(n):
-        return (1 - 1 / n, 1 / n)
+    def default_spi2_pn(n):
+        if n == 0:
+            return (0, 1)
+        else:
+            return (1 - 1 / (n+1), 1 / (n+1))
 
-    def solve_with_fp2(self, maxit=10, pn=default_fp2_pn):
+    def solve_with_spi2(self, maxit=10, pn=default_spi2_pn):
         self.diff_M = []
         # init
         QL_hat = np.copy(self.QL)
@@ -283,7 +289,7 @@ class MfgBase(object):
         # and average W in later iteration
         WL_bar = np.copy(self.QL)
         WR_bar = np.copy(self.QR)
-        for i in range(1, maxit + 1):
+        for i in range(0, maxit):
             # solve FP with Q
             Q = [self.QL, self.QR]
             self.solve_FP(Q)
@@ -291,7 +297,7 @@ class MfgBase(object):
             p1, p2 = pn(i)
             self.M_bar = p1 * self.M_bar + p2 * self.M
             # update Q_hat(firt compute W_bar)
-            if i == 1:
+            if p1 == 0:
                 WL_bar = self.M * self.QL
                 WR_bar = self.M * self.QR
                 QL_hat = np.copy(self.QL)  # avoid error here
@@ -319,7 +325,7 @@ class MfgBase(object):
         print("solving with fixed point method")
         self.diff_M = []
         self.diff_M_Mbar = []
-        for i in range(1, maxit + 1):
+        for i in range(0, maxit):
             Q = [self.QL, self.QR]
             self.solve_FP(Q)
             if fp:
@@ -382,6 +388,18 @@ class MfgBase(object):
         residual = np.sqrt(self.dx * self.dt) * norm(residual_)
         return residual
 
+    def compute_cost(self):
+        cost = 0.0
+        for ti in range(self.Nt):
+            m = self.M[ti, :]
+            u = self.U[ti, :]
+            ql = np.maximum((self.DL @ u) / self.F1_func(m), 0.0)
+            qr = np.minimum((self.DR @ u) / self.F1_func(m), 0.0)
+            q2 = ql**2 + qr**2
+            cost += np.sum(self.dt * self.dx * m * 0.5 * self.F1_func(m) * q2)
+        cost += np.sum(self.dx * self.UT * self.M[-1, :])
+        return cost
+
 
 class MfgBase2D(object):
     # it's a base class which implements the policy iteration and fixed point iteration
@@ -472,15 +490,15 @@ class MfgBase2D(object):
     # final condition
     # should return array with shape (self.Nx*self.Nx, )
     def UT_func(self, x1, x2):
-        raise NotImplementedError(self.__class__.__name__ + "didn't implement M0_func")
+        raise NotImplementedError(self.__class__.__name__ + "didn't implement UT_func")
 
     # hamilton
     # H(Du,m) = 0.5*|Du|^2/F1(m) - F2(m)
     def F1_func(self, m):
-        raise NotImplementedError(self.__class__.__name__ + "didn't implement M0_func")
+        raise NotImplementedError(self.__class__.__name__ + "didn't implement F1_func")
 
     def F2_func(self, m):
-        raise NotImplementedError(self.__class__.__name__ + "didn't implement M0_func")
+        raise NotImplementedError(self.__class__.__name__ + "didn't implement F2_func")
 
     def hjb_rhs(self, un, q, mn):
         ql1, qr1, ql2, qr2 = q
@@ -604,7 +622,10 @@ class MfgBase2D(object):
         return
 
     def default_fp_pn(n):
-        return (1 - 1 / n, 1 / n)
+        if n == 0:
+            return (0, 1)
+        else:
+            return (1 - 1 / (n+1), 1 / (n+1))
 
     def solve(self, maxit=10, fp=False, pn=default_fp_pn):
         # fp: bool, use fictious play or not
@@ -635,17 +656,17 @@ class MfgBase2D(object):
             self.logging_solution(self.M, self.U)
         return
 
-    def default_fp1_pn(n):
-        return (n / (1 + n), 1 / (1 + n))
+    def default_spi1_pn(n):
+        return (n / (n + 2), 1 / (n + 2))
 
-    def solve_with_fp1(self, maxit=10, pn=default_fp1_pn):
+    def solve_with_fp1(self, maxit=10, pn=default_spi1_pn):
         self.diff_M = []
         # init
         QL1_bar = np.copy(self.QL1)
         QR1_bar = np.copy(self.QR1)
         QL2_bar = np.copy(self.QL2)
         QR2_bar = np.copy(self.QR2)
-        for i in range(1, maxit + 1):
+        for i in range(0, maxit):
             # solve FP with Q_bar
             Q_bar = [QL1_bar, QR1_bar, QL2_bar, QR2_bar]
             self.solve_FP(Q_bar)
@@ -667,10 +688,13 @@ class MfgBase2D(object):
             self.logging_solution(self.M, self.U)
         return
 
-    def default_fp2_pn(n):
-        return (1 - 1 / n, 1 / n)
+    def default_spi2_pn(n):
+        if n == 0:
+            return (0, 1)
+        else:
+            return (1 - 1 / (n+1), 1 / (n+1))
 
-    def solve_with_fp2(self, maxit=10, pn=default_fp2_pn):
+    def solve_with_fp2(self, maxit=10, pn=default_spi2_pn):
         self.diff_M = []
         # init
         QL1_hat = np.copy(self.QL1)
@@ -685,7 +709,7 @@ class MfgBase2D(object):
         WR1_bar = np.copy(self.QR1)
         WL2_bar = np.copy(self.QL2)
         WR2_bar = np.copy(self.QR2)
-        for i in range(1, maxit + 1):
+        for i in range(0, maxit):
             # solve FP with Q
             Q = [self.QL1, self.QR1, self.QL2, self.QR2]
             self.solve_FP(Q)
@@ -693,7 +717,7 @@ class MfgBase2D(object):
             p1, p2 = pn(i)
             self.M_bar = p1 * self.M_bar + p2 * self.M
             # update Q_hat(firt compute W_bar)
-            if i == 1:
+            if p1 == 0:
                 WL1_bar = self.M * self.QL1
                 WR1_bar = self.M * self.QR1
                 WL2_bar = self.M * self.QL2
@@ -729,7 +753,7 @@ class MfgBase2D(object):
         print("solving with fixed point method")
         self.diff_M = []
         self.diff_M_Mbar = []
-        for i in range(1, maxit + 1):
+        for i in range(0, maxit):
             Q = [self.QL1, self.QR1, self.QL2, self.QR2]
             self.solve_FP(Q)
             if fp:
@@ -794,6 +818,20 @@ class MfgBase2D(object):
         residual = np.sqrt(self.dx**2 * self.dt) * norm(residual_)
         return residual
 
+    def compute_cost(self):
+        cost = 0.0
+        for ti in range(self.Nt):
+            m = self.M[ti, :]
+            u = self.U[ti, :]
+            ql1 = np.maximum((self.DL1 @ u) / self.F1_func(m), 0.0)
+            qr1 = np.minimum((self.DR1 @ u) / self.F1_func(m), 0.0)
+            ql2 = np.maximum((self.DL2 @ u) / self.F1_func(m), 0.0)
+            qr2 = np.minimum((self.DR2 @ u) / self.F1_func(m), 0.0)
+            q2 = ql1**2 + qr1**2 + ql2**2 + qr2**2
+            cost += np.sum(self.dt * self.dx**2 * m * 0.5 * self.F1_func(m) * q2)
+        cost += np.sum(self.dx**2 * self.UT * self.M[-1, :])
+        return cost
+
 
 def plot_residual(result, iterations: int):
     plt.plot(range(iterations), result.residual_FP_history, label="residual fp")
